# GP_EVAL.PY
# Dominic Ford
# 14/06/2006

# Evaluates expressions using user variables and functions

import sys
import re
import traceback
from math import *

from gp_error import *
import gp_spline

# GP_EVAL(): Evaluates an expression, substituting user defined variables and functions
def gp_eval(expression, vars, funcs, verbose=True, iteration=1):
  if (iteration > 20):
   raise OverflowError, "Iteration depth exceeded in function evaluation."
  try:
    # Evaluate any functions
    for func,fexp in funcs.iteritems():
      while 1:
        test = re.match(r"(.*)%s\s*\((.*)$"%func,expression)
        if (test == None): break
        bracketmatch = gp_bracketmatch(expression, test.start(2)-1)
        if (len(bracketmatch) == 0):            raise SyntaxError, "Mismatched brackets for function '%s'."%func
        if (len(bracketmatch) != abs(fexp[0])): raise SyntaxError, "Function '%s' takes %d arguments; %d provided."%(func,abs(fexp[0]),len(bracketmatch))
        args = []
        bracketmatch.insert(0,test.start(2)-1)
        for i in range(abs(fexp[0])):
          args.append(gp_eval(expression[ (bracketmatch[i]+1):bracketmatch[i+1] ], vars, funcs, False, iteration+1))
        if (fexp[0] < 0): # This is a spline
          try:
            value = gp_spline.spline_evaluate(args[0], fexp[1][0][1])
            expression = test.group(1) + str(value) + expression[bracketmatch[len(bracketmatch)-1]+1:]
          except:
            raise ValueError, "Error evaluating spline %s"%func
        else:             # This is a function
          funcdone = False
          for defno in range(len(fexp[1])):
            if not funcdone:
              j = len(fexp[1]) - 1 - defno
              func_scope = vars.copy()
              inrange = True
              for i in range(fexp[0]): func_scope[fexp[1][j][0][i]] = args[i]
              for i in range(fexp[0]):
                try:
                 if (fexp[1][j][1][i][0] != None): minrange = gp_eval(fexp[1][j][1][i][0],func_scope,funcs,False,iteration+1)
                 else                            : minrange = None
                 if (fexp[1][j][1][i][1] != None): maxrange = gp_eval(fexp[1][j][1][i][1],func_scope,funcs,False,iteration+1)
                 else                            : maxrange = None
                except:
                 if (verbose):
                  gp_error("Error evaluating range of function '%s'."%func)
                  gp_error("(it may be necessary to delete it with 'f(x)=' and then redefine it)")
                 raise
                if ((minrange != None) and (args[i] < minrange)): inrange = False
                if ((maxrange != None) and (args[i] > maxrange)): inrange = False
              if inrange:
                expression = test.group(1) + str(gp_eval(fexp[1][j][2],func_scope,funcs,False,iteration+1)) + expression[bracketmatch[len(bracketmatch)-1]+1:]
                funcdone   = True
          if not funcdone:
            raise ValueError, "Attempt to evaluate function '%s' with arguments out of their specified ranges."%func

    # All local functions substituted for; now evaluate what is left.
    return float(eval(expression, globals(), vars))
  except:
   if (verbose):
    gp_error("Error evaluating expression %s"%expression)
    gp_error("Error:" , sys.exc_info()[1], "(" , sys.exc_info()[0] , ")")
   raise

# GP_FUNCTION_DECLARE(): Declare a new function, possibly with a range
def gp_function_declare(line, funcs):
 test = re.match(r'([A-Za-z]\w*)\(([^()]*)\)\s*([^=]*)=(.*)',line)
 if (test == None):
  gp_error("Error: bad function definition '%s' could not be parsed."%line)
  return

 name       = test.group(1) # The name of the function
 arguments  = gp_split(test.group(2),",") # A textual list of its arguments
 arguments2 = [] # A stripped and checked list of its arguments
 ranges     = gp_split(test.group(3),"[")[1:] # List of range strings
 ranges2    = [] # A stripped and checked list of range strings
 expression = test.group(4).strip()

 for argument in arguments:
  test2 = re.match(r'^[A-Za-z]\w*$',argument.strip())
  if (test2 == None): raise SyntaxError, "Error: Function has badly formed argument name '%s'."%argument
  arguments2.append(argument.strip())

 for i in range(len(arguments)):
  if (i >= len(ranges)):
   ranges2.append([None, None]) # No range, if none specified
  else:
   rangestr = ranges[i].strip()
   if (rangestr == "]"): # Case [] means all x
    min = max = None
   else:
    test2 = re.match("([^:\]]*)((:)|( *to *))([^:\]]*)\]", rangestr)
    if (test2 == None): raise SyntaxError, "Error: range of argument %d should take form [min:max] or [min to max], but instead has form '[%s'."%(i+1,rangestr)
    min = test2.group(1).strip()
    max = test2.group(5).strip()
    if (len(min) == 0): min = None
    if (len(max) == 0): max = None
   ranges2.append([min,max])

 if   (len(expression) == 0)                                  :
  try:
   del funcs[name]
  except:
   raise KeyError, "Attempt to delete a function '%s' which does not exist."%name
 elif ((name in funcs) and (funcs[name][0] == len(arguments))): funcs[name][1].append([arguments2,ranges2,expression])
 else                                                         : funcs[name] = [len(arguments), [[arguments2,ranges2,expression]]]

# GP_BRACKETMATCH(): Find a matching closing bracket for an opening bracket

def gp_bracketmatch(expression, i):
  # Returns a list of comma positions in expression
  # Final item on list is closing )

  bracket_level = 0
  max           = len(expression)
  list          = []

  while( i < max-1 ):
    i = i+1 # Start searching after bracket at position i
    if (expression[i]==')'):
      if (bracket_level == 0):
        list.append(i)
        return list
      else:
        bracket_level = bracket_level - 1
    elif (expression[i]=='('):
        bracket_level = bracket_level + 1
    elif (expression[i]==','):
      if (bracket_level == 0):
        list.append(i)
        continue

# GP_SPLIT(): An intelligent string splitter, which doesn't grab split characters embedded in () or ""

def gp_split(expression, splitchar):
  bracket_level = 0
  quote_level   = 0
  apostro_level = 0
  max           = len(expression)
  list          = []
  i             = -1
  word          = ""

  while( i < max-1 ):
    i = i+1
    if   (expression[i]==')'): bracket_level = bracket_level - 1
    elif (expression[i]=='('): bracket_level = bracket_level + 1 
    elif ((expression[i]=='"') and (apostro_level == 0)): quote_level   = 1 - quote_level   # Hyphens allowed in quotes and vice-versa
    elif ((expression[i]=="'") and (quote_level   == 0)): apostro_level = 1 - apostro_level
    elif (expression[i]==splitchar):
      if ((bracket_level == 0) and (quote_level == 0) and (apostro_level == 0)):
        list.append(word)
        word=""
        continue
    word = word + expression[i]
  list.append(word)
  return list

# We've hit the end of string without brackets being closed
  return ()
